;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; $Id: hello.s,v 1.3 2000/09/03 07:01:53 cjh Exp $
;
; 68hc11 "Hello World" test program.
;
; Loads at 0x100 and used for testing db11's debug modes.
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
REGS		=	0x1000	; Device register location
SCCR2		=	0x2D	; SCI control register 2
SCSR		=	0x2E	; SCI status register
SCDR		=	0x2F	; SCI data register
RDRF		=	0x20	; SCSR "Received Data Ready" flag
TDRE		=	0x80	; SCSR "Transmit Data Register Empty" flag

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Definitions
STACK_BASE	=	0x200

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Startup code
Start:
		lds	#STACK_BASE-1	; Set stack pointer
		jsr	Delay229	; Startup delay

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
MainLoop:
		ldx	#HelloWorld	; Point to string
NextChar:	ldab	0,x
		beq	StartAgain	; End of string
		inx
		jsr	Delay229
		jsr	OUTSCI		; 
		jsr	INPOLL
		bra	NextChar
StartAgain:	bra	MainLoop

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Delay about .229 seconds (4Mhz). Zeroes Y
Delay229:	ldy	#0
Next229:	dey
		bne	Next229
		rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Put a character to SCI from B register, clobbers A
OUTSCI:
		ldaa	REGS+SCSR	; Wait for TDRE
		anda	#TDRE		; output ready?
		beq	OUTSCI		; not yet
		stab	REGS+SCDR	; send it, *SETTING CONDITION CODES*
		rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Poll for input character:
INPOLL:		ldab	REGS+SCSR	; Wait for RDRF
		andb	#RDRF		; character available?
		beq	INPOLL_EXIT	; not yet
		ldab	REGS+SCDR	; get it
		cmpb	#3		; ^C?
		bne	STARTAGAIN	; No, start output string again
		swi			; Programmed breakpoint
		bra	INPOLL_EXIT
STARTAGAIN:	ldx	#HelloWorld	; Point to string
INPOLL_EXIT:	rts

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Data
HelloWorld:	.asciz	"Hello, World\r\n"
