/*
 * $Id: hc11.h,v 1.1 2002/04/01 12:48:28 cjh Exp $
 *
 * MC68HC11 Header file
 *
 * Handles all series from A through K - define HC11F or HC11K if needed
 * Haven't checked differences between F0/F1, or K0/K1/K4.
 */
#ifndef	HC11_H
#define	HC11_H

typedef	unsigned char	uchar;	/* Not the right place for these but */
typedef	unsigned short	ushort;	/* they make the file stand alone */

#if	!defined(MC68HC11_REG_LOC)
#define	MC68HC11_REG_LOC 0x1000	/* Default location of registers */
#endif

/*
 * Define a pseudo-variable for the register block.
 * With this you can say "CPU.porta = 0;", for example
 */
#define	CPU	(*(CPU_t*)MC68HC11_REG_LOC)

typedef struct
{
#if	!defined(HC11K)
	uchar		baud;	/* 0x00,0x2B, baud rate register */
	volatile uchar	sccr1;	/* 0x01,0x2C, control reg 1 */
	uchar		sccr2;	/* 0x02,0x2D, control reg 2 */
	volatile uchar	scsr;	/* 0x03,0x2E, status register */
	volatile uchar	scdr;	/* 0x04,0x2F, data register */
#else
	uchar		scbdh;	/* 0x00,0x70 Baud rate Hi */
	uchar		scbdl;	/* 0x01,0x71, Baud rate Hi */
	volatile uchar	sccr1;	/* 0x02,0x72, control reg 1 */
	uchar		sccr2;	/* 0x03,0x73, control reg 2 */
	volatile uchar	scsr1;	/* 0x04,0x74, status register */
	volatile uchar	scsr2;	/* 0x05,0x75, status register */
	volatile uchar	scdrh;	/* 0x06,0x76, data register */
	volatile uchar	scdrl;	/* 0x07,0x77, data register */
#endif
} SCI_t;

typedef struct
{
	volatile uchar	porta;	/* +0x00, Port A Data */
#if	defined(HC11F)
	uchar		ddra;	/* +0x01, Port A Data Direction Register */
	volatile uchar	portg;	/* +0x02, Port G */
	uchar		ddrg;	/* +0x03, Port G Data Direction Register */
	volatile uchar	portb;	/* +0x04, Port B Data */
	volatile uchar	portf;	/* +0x05, Port F */
	volatile uchar	portc;	/* +0x06, Port C */
#elif	defined(HC11K)
	uchar		ddra;	/* +0x01, Port A Data Direction Register */
	uchar		ddrb;	/* +0x02, Port B Data Direction Register */
	uchar		ddrf;	/* +0x03, Port F Data Direction Register */
	volatile uchar	portb;	/* +0x04, Port B Data */
	volatile uchar	portf;	/* +0x05, Port F */
	volatile uchar	portc;	/* +0x06, Port C */
#else
	uchar		_res1;	/* +0x01 */
	volatile uchar	pioc;	/* +0x02, Parallel I/O Control */
	volatile uchar	portc;	/* +0x03, Port C Data */
	volatile uchar	portb;	/* +0x04, Port B Data */
	volatile uchar	portcl;	/* +0x05, Port C Latched */
	uchar		_res2;	/* +0x06 */
#endif
	uchar		ddrc;	/* +0x07, Port C Data Direction Register */
	volatile uchar	portd;	/* +0x08, Port D Data */
	uchar		ddrd;	/* +0x09, Port D Data Direction Register */
	volatile uchar	porte;	/* +0x0A, Port E Data */

	uchar		cforc;	/* +0x0B, Timer Compare Force */
	uchar		oc1m;	/* +0x0C, Output Compare 1 Mask */
	uchar		oc1d;	/* +0x0D, Output Compare 1 Data */
	volatile short	tcnt;	/* +0x0E, Timer Count */
	volatile short	tic1;	/* +0x10, Timer Input Capture 1 */
	volatile short	tic2;	/* +0x12, Timer Input Capture 2 */
	volatile short	tic3;	/* +0x14, Timer Input Capture 3 */
	short		toc1;	/* +0x16, Timer Output Compare 1 */
	short		toc2;	/* +0x18, Timer Output Compare 2 */
	short		toc3;	/* +0x1A, Timer Output Compare 3 */
	short		toc4;	/* +0x1C, Timer Output Compare 4 */
	volatile short	ti4o5;	/* +0x1E, Timer Input Capture 4, Out Compare 5*/
	uchar		tctl1;	/* +0x20, Timer Control 1 */
	uchar		tctl2;	/* +0x21, Timer Control 2 */
	uchar		tmsk1;	/* +0x22, Timer Interrupt Mask 1 */
	volatile uchar	tflg1;	/* +0x23, Timer Interrupt Flag 1 */
	uchar		tmsk2;	/* +0x24, Timer Interrupt Mask 2 */
	volatile uchar	tflg2;	/* +0x25, Timer Interrupt Flag 2 */
	uchar		pactl;	/* +0x26, Pulse Accumulator Control */
	volatile uchar	pacnt;	/* +0x27, Pulse Accumulator Count */

	uchar		spcr;	/* +0x28, Serial Peripheral Control */
	volatile uchar	spsr;	/* +0x29, Serial Peripheral Status */
	volatile uchar	spdr;	/* +0x2A, Serial Peripheral Data */

#if	!defined(HC11K)
	SCI_t		sci1;	/* +0x2B, SCI */
#else
	volatile uchar	eprog;	/* +0x2B, EPROG */
	volatile uchar	ppar;	/* +0x2C, */
	volatile uchar	pgar;	/* +0x2D, */
	uchar		_res7;	/* +0x2E, */
	uchar		_res8;	/* +0x2F, */
#endif

	volatile uchar	adctl;	/* +0x30, A/D Control/Status */
	volatile uchar	adr1;	/* +0x31, A/D Results 1 */
	volatile uchar	adr2;	/* +0x32, A/D Results 2 */
	volatile uchar	adr3;	/* +0x33, A/D Results 3 */
	volatile uchar	adr4;	/* +0x34, A/D Results 4 */

	uchar		bprot;	/* +0x35, EEPROM Block Protect */
	uchar		_res3;
#if	defined(HC11F)
	uchar		_res4;
	uchar		opt2;	/* +0x38, Second Option register */
#elif	defined(HC11K)
	uchar		init2;	/* +0x37, Second Init Register */
	uchar		opt2;	/* +0x38, Second Option register */
#else
	uchar		_res4;
	uchar		_res5;
#endif
	uchar		option;	/* +0x39, System Option Configurations */
	volatile uchar	coprst;	/* +0x3A, Arm/Reset COP Timer Circuitry */
	volatile uchar	pprog;	/* +0x3B, (E)EPROM Programming Control */
	uchar		hprio;	/* +0x3C, Highest Priority I-Bit Interrupt etc*/
	uchar		init;	/* +0x3D, RAM and I/O Mapping */
	uchar		test1;	/* +0x3E, Factory Test */
	const uchar	config;	/* +0x3F, (EEPROM) COP, ROM, EEPROM Enables */

#if	defined(HC11F)
	const uchar	_res6[0x1c];
	uchar		csstrh; /* +0x5C, chip select clock stretching */
	uchar		csctl;  /* +0x5D, chip select control */
	uchar		csgadr; /* +0x5E, chip select address register */
	uchar		csgsiz; /* +0x5F, chip select size register */
#elif	defined(HC11K)
	const uchar	_res6[0x15];
	uchar		mmsiz;	/* +0x56, */
	uchar		mmwbr;	/* +0x57, */
	uchar		mm1cr;	/* +0x58, */
	uchar		mm2cr;	/* +0x59, */
	uchar		cscstr;	/* +0x5A, */
	uchar		csctl;	/* +0x5B, */

	uchar		gpcs1a;	/* +0x5C, */
	uchar		gpcs1c;	/* +0x5D, */
	uchar		gpcs2a;	/* +0x5E, */
	uchar		gpcs2c;	/* +0x5F, */

	uchar		pwclk;	/* +0x60, PWM */
	uchar		pwpol;	/* +0x61, PWM */
	uchar		pwscal;	/* +0x62, PWM */
	uchar		pwen;	/* +0x63, PWM */
	uchar		pwcnt1;	/* +0x64, PWM */
	uchar		pwcnt2;	/* +0x65, PWM */
	uchar		pwcnt3;	/* +0x66, PWM */
	uchar		pwcnt4;	/* +0x67, PWM */
	uchar		pwper1;	/* +0x68, PWM */
	uchar		pwper2;	/* +0x69, PWM */
	uchar		pwper3;	/* +0x6A, PWM */
	uchar		pwper4;	/* +0x6B, PWM */
	uchar		pwdty1;	/* +0x6C, PWM */
	uchar		pwdty2;	/* +0x6D, PWM */
	uchar		pwdty3;	/* +0x6E, PWM */
	uchar		pwdty4;	/* +0x6F, PWM */

	SCI_t		sci1;	/* +0x70, SCI */

	uchar		_res7[4];
	volatile uchar	porth;	/* +0x7C, Port H */
	uchar		ddrh;	/* +0x7D, Port H Data Direction Register */
	volatile uchar	portg;	/* +0x7E, Port G */
	uchar		ddrg;	/* +0x7F, Port G Data Direction Register */
#endif
} CPU_t;

/* Bit names for general use */
#define	bit7	0x80
#define	bit6	0x40
#define	bit5	0x20
#define	bit4	0x10
#define	bit3	0x08
#define	bit2	0x04
#define	bit1	0x02
#define	bit0	0x01

/* PORTA bit definitions 0x00 */
#define	PA7	bit7
#define	PA6	bit6
#define	PA5	bit5
#define	PA4	bit4
#define	PA3	bit3
#define	PA2	bit2
#define	PA1	bit1
#define	PA0	bit0

/* PIOC bit definitions 0x02 */
#define	STAF	bit7
#define	STAI	bit6
#define	CWOM	bit5
#define	HNDS	bit4
#define	OIN	bit3
#define	PLS	bit2
#define	EGA	bit1
#define	INVB	bit0

/* PORTC bit definitions 0x03 */
#define	PC7	bit7
#define	PC6	bit6
#define	PC5	bit5
#define	PC4	bit4
#define	PC3	bit3
#define	PC2	bit2
#define	PC1	bit1
#define	PC0	bit0

/* PORTB bit definitions 0x04 */
#define	PB7	bit7
#define	PB6	bit6
#define	PB5	bit5
#define	PB4	bit4
#define	PB3	bit3
#define	PB2	bit2
#define	PB1	bit1
#define	PB0	bit0

/* PORTCL bit definitions 0x05 */
#define	PCL7	bit7
#define	PCL6	bit6
#define	PCL5	bit5
#define	PCL4	bit4
#define	PCL3	bit3
#define	PCL2	bit2
#define	PCL1	bit1
#define	PCL0	bit0

/* DDRC bit definitions 0x07 */
#define	DDC7	bit7
#define	DDC6	bit6
#define	DDC5	bit5
#define	DDC4	bit4
#define	DDC3	bit3
#define	DDC2	bit2
#define	DDC1	bit1
#define	DDC0	bit0

/* PORTD bit definitions 0x08 */
#define	PD5	bit5
#define	PD4	bit4
#define	PD3	bit3
#define	PD2	bit2
#define	PD1	bit1
#define	PD0	bit0

/* DDRD bit definitions 0x09 */
#define	DDD5	bit5
#define	DDD4	bit4
#define	DDD3	bit3
#define	DDD2	bit2
#define	DDD1	bit1
#define	DDD0	bit0

/* PORTE bit definitions 0x0A */
#define	PE7	bit7
#define	PE6	bit6
#define	PE5	bit5
#define	PE4	bit4
#define	PE3	bit3
#define	PE2	bit2
#define	PE1	bit1
#define	PE0	bit0

/* CFORC bit definitions 0x0B */
#define	FOC1	bit7
#define	FOC2	bit6
#define	FOC3	bit5
#define	FOC4	bit4
#define	FOC5	bit3

/* OC1M bit definitions 0x0C */
#define	OC1M7	bit7
#define	OC1M6	bit6
#define	OC1M5	bit5
#define	OC1M4	bit4
#define	OC1M3	bit3

/* OC1D bit definitions 0x0D */
#define	OC1D7	bit7
#define	OC1D6	bit6
#define	OC1D5	bit5
#define	OC1D4	bit4
#define	OC1D3	bit3

/* TCTL1 bit definition 0x20 */
#define	OM2	bit7
#define	OL2	bit6
#define	OM3	bit5
#define	OL3	bit4
#define	OM4	bit3
#define	OL4	bit2
#define	OM5	bit1
#define	OL5	bit0

/* TCTL2 bit definitions 0x21 */
#define	EDG4B	bit7		/* Capture on falling edge */
#define	EDG4A	bit6		/* Capture on rising edge */
#define	EDG1B	bit5		/* Capture on falling edge */
#define	EDG1A	bit4		/* Capture on rising edge */
#define	EDG2B	bit3		/* Capture on falling edge */
#define	EDG2A	bit2		/* Capture on rising edge */
#define	EDG3B	bit1		/* Capture on falling edge */
#define	EDG3A	bit0		/* Capture on rising edge */

/* TMSK1 bit definitions 0x22 */
#define	OC1I	bit7
#define	OC2I	bit6
#define	OC3I	bit5
#define	OC4I	bit4
#define	I4O5I	bit3
#define	IC1I	bit2
#define	IC2I	bit1
#define	IC3I	bit0

/* TFLG1 bit definitions 0x23 */
#define	OC1F	bit7
#define	OC2F	bit6
#define	OC3F	bit5
#define	OC4F	bit4
#define	I4O5F	bit3
#define	IC1F	bit2
#define	IC2F	bit1
#define	IC3F	bit0

/* TMSK2 bit definitions 0x24 */
#define	TOI	bit7
#define	RTII	bit6
#define	PAOVI	bit5
#define	PAII	bit4
#define	PR1	bit1
#define	PR0	bit0

/* TFLG2 bit definitions 0x25 */
#define	TOF	bit7
#define	RTIF	bit6
#define	PAOVF	bit5
#define	PAIF	bit4

/* PACTL bit definitions 0x26 */
#define	DDRA7	bit7
#define	PAEN	bit6
#define	PAMOD	bit5
#define	PEDGE	bit4
#define	DDRA3	bit3
#define	I4O5	bit2
#define	RTR1	bit1
#define	RTR0	bit0

/* SPCR bit definitions 0x28 */
#define	SPIE	bit7
#define	SPE	bit6
#define	DWOM	bit5
#define	MSTR	bit4
#define	CPOL	bit3
#define	CPHA	bit2
#define	SPR1	bit1
#define	SPR0	bit0

/* SPSR bit definitions 0x29 */
#define	SPIF	bit7
#define	WCOL	bit6
#define	MODF	bit4

/* BAUD bit definitions 0x2B */
#define	TCLR	bit7
#define	SCP1	bit5
#define	SCP0	bit4
#define	RCKB	bit3
#define	SCR2	bit2
#define	SCR1	bit1
#define	SCR0	bit0

/* SCCR1 bit definition 0x2C */
#define	R8	bit7
#define	T8	bit6
#define	M	bit4
#define	WAKE	bit3

/* SCCR2 bit definitions 0x2D */
#define	TIE	bit7
#define	TCIE	bit6
#define	RIE	bit5
#define	ILIE	bit4
#define	TE	bit3
#define	RE	bit2
#define	RWU	bit1
#define	SBK	bit0

/* SCSR	 bit definitions 0x2E */
#define	TDRE	bit7
#define	TC	bit6
#define	RDRF	bit5
#define	IDLE	bit4
#define	OR	bit3
#define	NF	bit2
#define	FE	bit1

/* SCDR bit definitions 0x2F */
#define	R7T7	bit7
#define	R6T6	bit6
#define	R5T5	bit5
#define	R4T4	bit4
#define	R3T3	bit3
#define	R2T2	bit2
#define	R1T1	bit1
#define	R0T0	bit0

/* ADCTL bit definitions 0x30 */
#define	CCF	bit7
#define	SCAN	bit5
#define	MULT	bit4
#define	CD	bit3
#define	CC	bit2
#define	CB	bit1
#define	CA	bit0

/* BPROT bit definitions 0x35 */
#define	PTCON	bit4
#define	BPRT3	bit3
#define	BPRT2	bit2
#define	BPRT1	bit1
#define	BPRT0	bit0

/* OPTION bit definitions 0x39 */
#define	ADPU	bit7
#define	CSEL	bit6
#define	IRQE	bit5
#define	DLY	bit4
#define	CME	bit3
#define	CR1	bit1
#define	CR0	bit0

/* PPROG bit definitions 0x3B */
#define	ODD	bit7
#define	EVEN	bit6
#define	ELAT	bit5	/* MC68HC711E9 only */
#define	BYTE	bit4
#define	ROW	bit3
#define	ERASE	bit2
#define	EELAT	bit1
#define	EEPGM	bit0

/* HPRIO bit definitions 0x3C */
#define	RBOOT	bit7
#define	SMOD	bit6
#define	MDA	bit5
#define	IRVNE	bit4
#define	PSEL3	bit3
#define	PSEL2	bit2
#define	PSEL1	bit1
#define	PSEL0	bit0

/* INIT	 bit definitions 0x3D */
#define	RAM3	bit7
#define	RAM2	bit6
#define	RAM1	bit5
#define	RAM0	bit4
#define	REG3	bit3
#define	REG2	bit2
#define	REG1	bit1
#define	REG0	bit0

/* TEST1 bit definitions 0x3E */
#define	TILOP	bit7
#define	OCCR	bit5
#define	CBYP	bit4
#define	DISR	bit3
#define	FCM	bit2
#define	FCOP	bit1
#define	TCON	bit0

/* CONFIG bit definitions 0x3F */
#define	NOSEC	bit3
#define	NOCOP	bit2
#define	ROMON	bit1	/* MC68HC11E9 and MC68HC11E8 only */
#define	EPON	bit1	/* MC68HC711E9 only */
#define	EEON	bit0

/*
 * Interrupt enabling and disabling
 */
#define	ei()	asm("cli")
#define	di()	asm("sei")

/*
 * Bootstrap-mode vectors.
 */
typedef struct
{
	uchar	jmp;	/* hex of "jmp", 0x7E */
	ushort	addr;	/* address to jump to */
} BootVector;

#define	SCIBootVector	(*(BootVector*)0xC4)	/* SCI */
#define	SPIBootVector	(*(BootVector*)0xC7)	/* SPI */
#define	PAICBootVector	(*(BootVector*)0xCA)	/* Pulse Acc. Input Edge */
#define	PAOBootVector	(*(BootVector*)0xCD)	/* Pulse Acc. Overflow */
#define	TOVBootVector	(*(BootVector*)0xD0)	/* Timer Overflow */
#define	OC5BootVector	(*(BootVector*)0xD3)	/* Timer Output Compare 5 */
#define	OC4BootVector	(*(BootVector*)0xD6)	/* Timer Output Compare 4 */
#define	OC3BootVector	(*(BootVector*)0xD9)	/* Timer Output Compare 3 */
#define	OC2BootVector	(*(BootVector*)0xDC)	/* Timer Output Compare 2 */
#define	OC1BootVector	(*(BootVector*)0xDF)	/* Timer Output Compare 1 */
#define	IC3BootVector	(*(BootVector*)0xE2)	/* Timer Input Capture 3  */
#define	IC2BootVector	(*(BootVector*)0xE5)	/* Timer Input Capture 2  */
#define	IC1BootVector	(*(BootVector*)0xE8)	/* Timer Input Capture 1  */
#define	RTIBootVector	(*(BootVector*)0xEB)	/* Real-Time Interrupt  */
#define	IRQBootVector	(*(BootVector*)0xEE)	/* IRQ */
#define	XIRQBootVector	(*(BootVector*)0xF1)	/* XIRQ  */
#define	SWIBootVector	(*(BootVector*)0xF4)	/* SWI  */
#define	ILLBootVector	(*(BootVector*)0xF7)	/* Illegal Opcode  */
#define	COPBootVector	(*(BootVector*)0xFA)	/* COP Fail  */
#define	ClockMonBootVector (*(BootVector*)0xFD)	/* Clock Monitor Fail  */

static inline void
BootVectorInit(BootVector* bv, void (*addr)(void))
{
	bv->jmp = 0x7E;
	bv->addr = (ushort)addr;
}

#endif	/* HC11_H */
