/*
 * $Id: sciintr.c,v 1.1 2002/04/01 12:48:29 cjh Exp $
 *
 * Boot-mode program showing the use of SCI interrupts
 */
#include	<hc11.h>

/*
 * Configuration variables, (un)comment at will
 */
/* #define TRANSMIT_ON_COMPLETE	 1	/* Use TC instead of TDRE */
#define SHOW_OCTAL	1		/* Echo in octal */
#define RECV_INTR	1		/* Use receiver interrupts */

#if	defined(TRANSMIT_ON_COMPLETE)
#define	TRANSMIT_READY		TC
#define	TRANSMIT_INTR_EN	TCIE
#else
#define	TRANSMIT_READY		TDRE
#define	TRANSMIT_INTR_EN	TIE
#endif

/*
 * Miniature replacement for crt0:
 */
asm("
	.sect	.page0
	lds	#0x1FF
	jmp	main
");

/* Start SCI output by enabling interrupts */
inline static void SCIStart() { CPU.sci1.sccr2 |= TRANSMIT_INTR_EN; }

/* Stop SCI output by disabling interrupts */
inline static void SCIStop() { CPU.sci1.sccr2 &= ~TRANSMIT_INTR_EN; }

/* Return non-zero if SCI has input waiting */
inline static uchar SCIInputAvail() { return CPU.sci1.scsr & RDRF; }

/* Return input byte from SCI */
inline static uchar SCIInput() { return CPU.sci1.scdr; }

/* Transmit byte through SCI */
inline static uchar SCIOutputReady() { return CPU.sci1.scsr & TRANSMIT_READY; }

/* Transmit byte through SCI */
inline static void SCIOutput(uchar byte) { CPU.sci1.scdr = byte; }

inline static void
SCIInitBoot(void __attribute__((interrupt)) (*handler)())
{
	SCIBootVector.jmp = 0x7E;
	SCIBootVector.addr = (ushort)handler;

// Already done by DB11's talker, leave it alone:
//	CPU.sci1.sccr1 = R8|T8;		/* 8 bits, one stop bit */
//	CPU.baud = 0x30;		/* With 8MHz crystal, baud=9600 */

#if	defined(RECV_INTR)
	CPU.sci1.sccr2 = TRANSMIT_INTR_EN|RIE|TE|RE;
#else
	CPU.sci1.sccr2 = TRANSMIT_INTR_EN|TE|RE;
#endif
}

volatile uchar	input;		/* Single-character input buffer */
volatile uchar*	output;		/* Pointer to output data */
static char	outbuf[6];

/*
 * Handle the SCI interrupt
 */
void __attribute__((interrupt))
SCIHandler(void)
{
#if	defined(RECV_INTR)
	if (SCIInputAvail())
		input = SCIInput();	/* Read input */
#endif

	// This should be smaller code than below, but it doesn't work!?
	while (SCIOutputReady())
	{
		if (!output || *output == '\0')
		{
			SCIStop();	// Block further transmit interrupts
			output = 0;	// End of output
			break;
		}

		SCIOutput(*output++);
	}
}

uchar
SCIGet()
{
#if	defined(RECV_INTR)
	uchar	c;
	while (input == 0)		/* No need to di()/ei(), single byte */
		;
	if (input == 3)
		asm("swi");
	c = input;
	input = 0;
	return c;
#else
	while (!SCIInputAvail())
		;
	input = SCIInput();
	if (input == 3)
		asm("swi");
	return input;
#endif
}

/*
 * Wait for output to drain, checking "output" pointer atomically.
 * The C compiler can't find a sensible way to compile this, so it's assembly.
 */
void
__attribute__((section("page0")))
SCIDrain()
{
asm("
1:	sei
	ldab	output
	orab	output+1
	cli
	bne	1b"
);
}

void
__attribute__((section("page0")))
SCIPut(uchar c)
{
	SCIDrain();		/* Block until existing output has gone */

	while (!SCIOutputReady())
		;

	outbuf[0] = c;
	outbuf[1] = '\0';
	output = outbuf;
	SCIOutput(*output++);	/* Shouldn't be necessary! */
	SCIStart();		/* Enable interrupts */
}

#if defined(SHOW_OCTAL)
void
putoctal(unsigned short d)
{
	SCIDrain();
	di();
	output = outbuf+sizeof(outbuf)-1;
	*output = '\0';

	do *--output = '0'|(d&0x7);
	while (d >>= 3);
	ei();

	SCIOutput(*output++);	/* Start first output char */
	SCIStart();		/* Enable SCI transmitter interrupts */
}
#endif

int
__attribute__((section("page0")))
main(void) 
{
	di();			/* Disable interrupts */
	SCIInitBoot(SCIHandler);/* Setup SCI */
	input = 0;
	output = 0;
	ei();			/* Enable interrupts */

	for (;;)
	{
		uchar c = SCIGet();
#if defined(SHOW_OCTAL)
		putoctal(c);
		SCIPut(' ');
#else
		SCIPut(c);
#endif
	}
} 
