/*
 * $Id: break.cxx,v 1.6 2002/03/13 10:17:17 cjh Exp $
 *
 * Breakpoint table API for db11
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<util.h>
/* #include	<db11.h> */
#include	<serial.h>
#include	<comms.h>
#include	<talker.h>
#include	<break.h>

/*
 * Breakpont table
 */
enum BreakStatus {
	NotUsed,
	Disabled,
	Enabled
};
typedef struct
{
	ushort		address;	// What address is the breakpoint at?
	BreakStatus	status;		// Is the breakpoint set?
	boolean		have_set;	// If we've set it, we know the opcode
	uchar		opcode;		// What opcode was replaced by SWI?
} Breakpoint;

Breakpoint*	breaks;
int		num_break;
int		break_alloc;

// Add a breakpoint
int
BreakAdd(ushort addr)
{
	boolean		have_set = FALSE;
	uchar		opcode = 0;

	/*
	 * Check for existing breakpoint at same address
	 */
	for (int i = 0; i < num_break; i++)
	{
		if (breaks[i].status == NotUsed)
			continue;
		if (breaks[i].address == addr
		 && breaks[i].have_set)
		{
			//printf("Breakpoint %d is also at 0x%04X\n", i, addr);

			/* Snarf opcode from duplicate, not CPU */
			have_set = TRUE;
			opcode = breaks[i].opcode;
			break;
		}
	}

	if (num_break == break_alloc)
	{
		// Allocate more memory
		int		new_alloc = (break_alloc|0xF)+1;
		Breakpoint*	np = new Breakpoint[new_alloc];

		memcpy(np, breaks, sizeof(Breakpoint)*num_break);
		delete breaks;
		breaks = np;
		break_alloc = new_alloc;
	}
	breaks[num_break].address = addr;
	breaks[num_break].status = Disabled;
	breaks[num_break].have_set = have_set;
	breaks[num_break].opcode = opcode;
	num_break++;

	if (BreakEnableQuiet(num_break-1, TRUE))
		return num_break-1;
	return -1;
}

// Remove a breakpoint
boolean
BreakRemove(int brknum)
{
	if (brknum < 0 || brknum >= num_break)
	{
		printf("No such breakpoint %d\n", brknum);
		return FALSE;
	}
	if (breaks[brknum].status == Enabled)
		BreakEnableQuiet(brknum, FALSE);
	breaks[brknum].status = NotUsed;

	// Discount trailing unused breakpoints
	while (num_break > 0 && breaks[num_break-1].status == NotUsed)
		num_break--;

	return TRUE;
}

// List all breakpoints
void
BreakList()
{
	for (int i = 0; i < num_break; i++)
	{
		if (breaks[i].status == NotUsed)
			continue;
		printf("Breakpoint %d at 0x%04X is %sabled\n",
			i,
			breaks[i].address,
			breaks[i].status == Enabled ? "en" : "dis"
		);
	}
}

// Number of max breakpoint
int
BreakCount()
{
	return num_break;
}

boolean
BreakEnableQuiet(int brknum, boolean enable)
{
	if (brknum < 0 || brknum >= num_break)
	{
		printf("No such breakpoint %d\n", brknum);
		return FALSE;
	}

	if ((breaks[brknum].status == Enabled) == enable)
		return TRUE;	// Already ok

	if (enable)
	{
		// Get opcode we're replacing

		if (breaks[brknum].have_set)
		{
			// Already have the opcode
		}
		else if (!TalkerSetAddr(breaks[brknum].address)
		      || TalkerGetMem(&breaks[brknum].opcode, 1) != 1)
		{
			printf("Can't get opcode to set breakpoint\n");
			return FALSE;
		}
		breaks[brknum].have_set = TRUE;

		// Set SWI opcode (0x3F)
		if (TalkerWrite(breaks[brknum].address, (uchar*)"\x3F", 1, TRUE) != 1)
		{
			printf("Can't set breakpoint\n");
			return FALSE;
		}

		breaks[brknum].status = Enabled;
	}
	else
	{
		breaks[brknum].status = Disabled;
		if (BreakSetAt(breaks[brknum].address) >= 0)
		{
			return TRUE;	// Duplicate still set
		}
		// Reset original opcode
		if (TalkerWrite(
				breaks[brknum].address,
				&breaks[brknum].opcode,
				1,
				TRUE
			) != 1
		)
		{
			printf("Can't disable breakpoint\n");
			return FALSE;
		}

	}
	return TRUE;
}

boolean
BreakEnable(int brknum, boolean enable)
{
	boolean	ok = BreakEnableQuiet(brknum, enable);

	if (ok)
		printf("%sabled breakpoint %d at 0x%04X\n",
			breaks[brknum].status == Enabled ? "En" : "Dis",
			brknum,
			breaks[brknum].address
		);
	return ok;
}

// Which breakpoint is here?
int
BreakSetAt(ushort addr)
{
	for (int i = 0; i < num_break; i++)
	{
		if (breaks[i].status == Enabled
		 && breaks[i].address == addr)
			return i;
	}
	return -1;
}

// Which opcode was replaced; only valid while breakpoint is enabled
uchar
BreakOpCode(int brknum)
{
	if (brknum < 0 || brknum >= num_break)
	{
		printf("No such breakpoint %d\n", brknum);
		return FALSE;
	}
	if (!breaks[brknum].have_set)
	{
		printf("Internal error, opcode as-yet unknown for breakpoint\n");
		return 0x3F;	// SWI
	}
	return breaks[brknum].opcode;
}

ushort
BreakAddr(int brknum)
{
	if (brknum < 0 || brknum >= num_break)
	{
		printf("No such breakpoint %d\n", brknum);
		return 0;
	}
	return breaks[brknum].address;
}
