/*
 * $Id: comms.cxx,v 1.15 2002/06/10 03:26:12 cjh Exp $
 *
 * Download code to a 68hc11 and talk to it in various ways
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<util.h>
#include	<comms.h>
#include	<boot.h>
#include	<midi.h>
#include	<db11.h>
#include	<serial.h>
#include	<talker.h>
#include	<debug.h>

void
Connect(int baud)
{
	Notify("Connecting at %d baud...\n", baud);
	if (not_really)
		return;

	mcu->setSpeed(baud);

	if (DEBUG_MODE)
	{
		RunDebugger();
		return;
	}

//	stdio->disableLineEdit();	// Set keyboard to raw mode
	stdio->disableSpecialInput();
	ProcessInput();
}

void
ProcessInput()
{
	Multiplexed*	mpx_pp[2];
	boolean		done = FALSE;

	while (!done)
	{
		mpx_pp[0] = stdio;
		mpx_pp[1] = mcu;

		if (!Multiplex(mpx_pp, 2))
			continue;

		/* Process all data received: */
		done = FALSE;

		if (mcu->isReady()
		 && !processMCUInput())
		 	done = TRUE;

		if (stdio->isReady()
		 && !processKeyInput())
		 	done = TRUE;
	}
}

boolean
processMCUInput()
{
	unsigned char	buf[256];
	int		r;
	int		i;

	r = mcu->read(buf, sizeof(buf));
	for (i = 0; i < r; i++)
		Dump((char)buf[i]);
	return TRUE;
}

boolean
processKeyInput()
{
	unsigned char	buf[256];
	int		r = stdio->read(buf, 1); // Read data from the keyboard
	static boolean	after_newline = FALSE;

	/*
	 * Detect a request to enter a local commands, process it and return
	 */
	if (buf[0] == '~' && after_newline)
	{
		const unsigned char* prompt = (const unsigned char*)"command? ";
		stdio->disableSpecialInput(FALSE);	// Re-enable
		stdio->write(prompt, strlen((char*)prompt));
		r = stdio->read(buf, sizeof(buf));
		if (buf[0] != '~')	/* an escaped tilde! */
		{
			boolean more = ProcessCommandLine((char*)buf);
			stdio->disableSpecialInput();
			return more;
		}
	}

	// Catch ^Z as an interrupt to connect mode.
	if (buf[0] == '\x1A')
		return FALSE;

	after_newline = buf[0] == '\n' || buf[0] == '\r';

	if (LOCAL_ECHO)
		DumpAscii(buf[0]);

	Send(buf[0]);	/* Only asked for 1 byte */
	return TRUE;
}

/*
 * Process a ~ escaped db11 command line.
 * Returns 0 for failure, 1 to continue processing the char
 *
 * REVISIT: This should simply use the db11 Command() processor.
 */
boolean
ProcessCommandLine(char* buf)
{
	char*		cp;

	/*
	 * Remove trailing newline characters.
	 */
	cp = buf+strlen(buf);
	while (--cp >= buf && (*cp == '\r' || *cp == '\n'))
		*cp-- = '\0';

	switch (buf[0])
	{
	case 'E':	/* Toggle local echo mode */
	case 'e':
		LOCAL_ECHO = !LOCAL_ECHO;
		DescribeSendMode();
		break;

	case 'H':	/* Toggle Hex output mode */
	case 'h':
		HEX_OUTPUT = !HEX_OUTPUT;
		DescribeDumpMode();
		break;

	case 'I':	/* Toggle Hex input mode */
	case 'i':
		HEX_INPUT = !HEX_INPUT;
		DescribeSendMode();
		break;

	case 'M':	/* Toggle MIDI mode */
	case 'm':
		MIDI_MODE = !MIDI_MODE;
		DescribeDumpMode();
		break;

	case 'Q':	/* Quit connect mode */
	case 'q':
		stdio->disableSpecialInput(FALSE);		// Re-enable
		printf("\nBye\n");
		return FALSE;

#if 0	/* Disabled because we don't have "fd" in here */
	case '!':	/* Run an external program */
		/*
		 * This command is intended to allow
		 * hooking in an external program that
		 * knows how to communicate in a special
		 * serial protocol, e.g. a debugger etc.
		 *
		 * The standard input is connected to
		 * the mc68hc11, standard output must
		 * be used to access the keyboard and
		 * screen.
		 */
		sprintf(cmd, "sh -c '%s' <&%d", buf+1, fd);

		stdio->suspend();

#if	defined(SIGTSTP)
		signal(SIGTSTP, (sighandler_t)SIG_DFL);
#endif	/* SIGTSTP */

		system(cmd);	/* Execute the command */

#if	defined(SIGTSTP)
		signal(SIGTSTP, (sighandler_t)SuspendMe);
#endif	/* SIGTSTP */

		stdio->resume();
		printf("!\r\n");
		break;
#endif

	case '?':	/* help! */
		printf("Tilde commands:\r\n"
			"e\ttoggle local echo\r\n"
			"hi\ttoggle hex input mode\r\n"
			"ho\ttoggle hex output mode\r\n"
			"m\ttoggle MIDI mode\r\n"
#if 0
			"! cmd\trun external cmd\r\n"
#endif
			"~\tsend a tilde\r\n"
			"q\tquit\r\n"
		);
		break;

	default:
		printf("Unrecognised command '%s'\n", buf);
		break;
	}
	return TRUE;
}

/*
 * Looking at mode flags, decide how to send this byte to the MCU.
 */
void
Send(char ch)
{
	if (HEX_INPUT)
	{
		SendHex(ch);
		return;
	}
	SendDirect(ch);
}

/*
 * Looking at mode flags, decide how to display this byte on the screen.
 */
void
Dump(char ch)
{
	if (MIDI_MODE)
	{
		if (HEX_OUTPUT)
			DumpHex(ch);
		DumpMidi(ch);
		return;
	}
	if (HEX_OUTPUT)
	{
		DumpHex(ch);
		return;
	}
	DumpAscii(ch);
}

/*
 * Write a line describing the new input mode
 */
void
DescribeSendMode()
{
/*
	int	expect_echo = FALSE;

	if (send == SendDirect)
		printf("Direct");
	else if (send == SendHex)
	{
		expect_echo = TRUE;
		printf("Hex");
	}
	printf(" input mode");
	if (expect_echo != local_echo)
		printf(" with%s local echo", local_echo ? "" : "out");
	printf("\n");
*/
}

/*
 * Send function to write character directly
 */
void
SendDirect(char ch)
{
	mcu->write((unsigned char*)&ch, 1);
}

/*
 * Send function to gather hex bytes and send them (each 2nd char sends a byte)
 */
void
SendHex(char ch)
{
	static	int hex_have_1st = FALSE;/* Have we the first nybble of hex? */
	static	int	hex_nybble = 0;	/* The nybble */
	const char*	cp;

	if (ch == 0 || ch == '\b')
	{
		if (ch == '\b' && hex_have_1st)
			/* Erase typed character */
			stdio->write((const unsigned char*)"\b \b", 3);
		hex_have_1st = FALSE;
		return;
	}

	ch = toupper(ch);
	if (!(cp = strchr(hex, ch)))
	{	/* Not hex, reset! */
		hex_have_1st = FALSE;
		stdio->write((const unsigned char*)"\7 not hex!\r\n", 12);
	}
	else if (!hex_have_1st)
	{	/* first hex nybble */
		hex_have_1st = TRUE;
		hex_nybble = cp-hex;
	}
	else
	{	/* second hex nybble */
		hex_have_1st = FALSE;
		ch = cp-hex + (hex_nybble<<4);
		mcu->write((unsigned char*)&ch, 1);	/* Write byte */
		stdio->write((const unsigned char*)"\r\n", 2); /* Echo newline */
	}
}

/*
 * Write a line describing the new output mode:
 */
void
DescribeDumpMode()
{
/*
	if (dump == DumpAscii)
		printf("Ascii");
	else if (dump == DumpHex)
		printf("Hex");
	else if (dump == DumpMidi)
		printf("Midi");
	else if (dump == DumpHexAndMidi)
		printf("Midi with Hex");
	printf(" output mode\n");
*/
}

/*
 * Display received character as hex and binary
 */
void
DumpHex(char ch)
{
	int	j;
	printf("x%02X ", ch&0xFF);
	for (j = 0x80; j; j >>= 1)
		putchar('0'+(ch&j?1:0));
	putchar('\n');
	fflush(stdout);
}

/*
 * Present an ascii-readable version of the character.
 */
void
DumpAscii(char ch)
{
	printf("%s", Ascii(ch));
	fflush(stdout);
}
