/*
 * $Id: debug.cxx,v 1.33 2002/04/01 13:02:14 cjh Exp $
 *
 * Command-line mode for the Talker.
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>

#include	<ctype.h>

#include	<util.h>
#include	<comms.h>
#include	<db11.h>
#include	<serial.h>
#include	<talker.h>
#include	<debug.h>
#include	<disasm.h>
#include	<break.h>
#include	<symbol.h>

static void	DebugRead(char**& argv);
static void	DebugZero(char**& argv);
static void	DebugLoadS19(char**& argv);
static void	DebugRegisters(char**& argv);
static void	DebugSet(char**& argv);
static void	DebugBreakpoint(char**& argv);
static void	DebugUnbreak(char**& argv);
static void	DebugEnable(char**& argv, boolean enable);
static void	DebugContinue(char**& argv);
static void	DebugDisassemble(char**& argv);
static void	DebugSingleStep(boolean over);
static void	DebugVerify(char**& argv);

static void	PrintRegisters(Frame* regs);
static void	Disassemble(ushort start, ushort finish);
static int	DisGetByte(void* closure);
static void	DisUngetByte(void* closure, char ch);
static ushort	NextInstruction(Frame* regs, ushort* over_call);
static boolean	SingleStep(Frame* regs, boolean over);

static int	stopped_at = -1;	// Which breakpoint are we stopped at?

/*
 * This is the talker-mode interactive command interpreter:
 */
void
RunDebugger()
{
	uchar		cmd[257];
	int		r;
	char**		argv = 0;

	printf("Type h<ENTER> for help\n");
	for (;;)
	{
		/* Prompt */
		stdio->write((uchar*)"Talk> ", 6);
		if ((r = stdio->read(cmd, 256)) < 0)
		{
			if (errno != EINTR)
			{
				strcmdfree(argv);
		 		return;
			}
			else
				continue;
		}

		cmd[r] = '\0';
		int	argc;
		strcmdfree(argv);
		argv = strcmd((char*)cmd, ",=@#;", &argc);

		if (!CommandList(argv, DebugCommand))
			break;		// Debugger is done.
	}
	strcmdfree(argv);
}
/*
 * Run a single debug command
 */
boolean
DebugCommand(char**& argv)
{

	if (!argv || !*argv)
		return TRUE;

#if 0
printf("DebugCommand(");
for (int i = 0; argv[i]; i++)
	printf("%s%s", i?" ":"", argv[i]);
printf(")\n");
#endif

	char*	cmd;
	switch (tolower(*(cmd = *argv++)))
	{
	case 'b':
		if (strcasecmp(cmd, "break") == 0
		 || strcasecmp(cmd, "b") == 0)
		{
			DebugBreakpoint(argv);
			return TRUE;
		}
		break;

	case 'c':			// Continue program execution
		if (strcasecmp(cmd, "continue") == 0
		 || strcasecmp(cmd, "c") == 0)
		{
			DebugContinue(argv);
			return TRUE;
		}
		break;

	case 'd':			// Disable breakpoint
		if (strcasecmp(cmd, "disable") == 0
		 || strcasecmp(cmd, "d") == 0)
		{
			DebugEnable(argv, FALSE);
			return TRUE;
		}
		else if (strcasecmp(cmd, "del") == 0
		      || strcasecmp(cmd, "delete") == 0)
		{
			DebugUnbreak(argv);
			return TRUE;
		}
		break;

	case 'e':			// Enable breakpoint
		if (strcasecmp(cmd, "enable") == 0
		 || strcasecmp(cmd, "e") == 0)
		{
			DebugEnable(argv, TRUE);
			return TRUE;
		}
		break;

	case 'f':			// Load an S19 file
		if (strcasecmp(cmd, "file") == 0
		 || strcasecmp(cmd, "f") == 0)
		{
			DebugLoadS19(argv);
			return TRUE;
		}
		break;

	case 'g':			// Read and display memory
		if (strcasecmp(cmd, "get") == 0
		 || strcasecmp(cmd, "g") == 0)
		{
			DebugRead(argv);
			return TRUE;
		}
		break;

	case 'h':
	case '?':
		printf(
			"Talker-mode commands (literal values are C-style):\n"
			"\n"
		//	"a[ssemble]\t\t\tAssemble into memory (not yet)\n"
			"b[reak] addr | l\t\tSet breakpoint at address or list all\n"
			"c[ontinue]\t\t\tRun program from current location\n"
			"d[isable] [num]\t\t\tDisable [specified] breakpoint\n"
			"e[nable] [num]\t\t\tEnable [specified] breakpoint\n"
			"f[ile] file\t\t\tLoad an S19 file into memory\n"
			"g[et] [addr][,count|-addr2]\tDisplay memory\n"
			"l[ist] [addr][,count]\t\tDisassemble code\n"
			"n[ext]\t\t\t\tExecute next instruction\n"
			"o[ver]\t\t\t\tNext, step over subroutine call\n"
			"q[uit]\t\t\t\tQuit db11\n"
			"r[eg]\t\t\t\tDisplay register frame\n"
			"s[et] [a|b|ccr|d|pc|sp|x|y|*addr]=value\tSet register or memory\n"
			"v[erify] [addr][,count|-addr2]\tVerify memory\n"
			"z[ero] [addr][,count|-addr2]\tZero memory\n"
		);
		return TRUE;

	case 'l':
		if (strcasecmp(cmd, "list") == 0
		 || strcasecmp(cmd, "l") == 0)
		{
			DebugDisassemble(argv);
			return TRUE;
		}
		break;

	case 'n':
		if (strcasecmp(cmd, "next") == 0
		 || strcasecmp(cmd, "n") == 0)
		{
			DebugSingleStep(FALSE /*, argv*/);
			return TRUE;
		}
		break;

	case 'o':
		if (strcasecmp(cmd, "over") == 0
		 || strcasecmp(cmd, "o") == 0)
		{
			DebugSingleStep(TRUE /*, argv*/);
			return TRUE;
		}
		break;

	case 'q':
		if (strcasecmp(cmd, "quit") == 0
		 || strcasecmp(cmd, "q") == 0)
		{
			return FALSE;
		}
		break;

	case 'r':			// Display registers
		if (strcasecmp(cmd, "reg") == 0
		 || strcasecmp(cmd, "r") == 0)
		{
			DebugRegisters(argv);
			return TRUE;
		}
		break;

	case 's':			// Set registers
		if (strcasecmp(cmd, "s") == 0
		 || strcasecmp(cmd, "set") == 0)
		{
			DebugSet(argv);
			return TRUE;
		}
		break;

	case 'v':			// Verify memory
		if (strcasecmp(cmd, "verify") == 0
		 || strcasecmp(cmd, "v") == 0)
		{
			DebugVerify(argv);
			return TRUE;
		}
		break;

	case 'z':
		if (strcasecmp(cmd, "zero") == 0
		 || strcasecmp(cmd, "z") == 0)
		{
			DebugZero(argv);
			return TRUE;
		}
		break;

	default:
		break;
	}

	/*
	 * Assume it's a normal db11 command.
	 */
	argv--;		// We incremented it.
	Command(argv);

	return TRUE;
}

/*
 * Interactive read command handler.
 * Read memory and display a hex and ASCII dump.
 */
static	void
DebugRead(char**& argv)
{
	long		start = eaddr, count = 16;
	uchar		buf[16];
	uchar		last[16];
	int		prev = 0;	// 0 = no buffer, 1 = buf, 2 = repeating

	GetAddressRange(argv, &start, &count);

	/*
	 * Don't set the address if it's set already.
	 */
	if (eaddr != start)
		TalkerSetAddr(start);

	/*
	 * Read memory in 16-byte blocks:
	 */
	while (count > 0)
	{
		int		thiscount;
		int		r, i;
		short		addr;

		thiscount = count;
		if (thiscount > 16)
			thiscount = 16;

		addr = eaddr;
		r = TalkerGetMem(buf, thiscount);

		if (prev
		 && thiscount == 16
		 && memcmp(last, buf, 16) == 0)
		{
			if (prev == 1)
				printf("*\n");
			prev = 2;
		}
		else
		{
			printf("%04X\t", addr&0xFFFF);
			for (i = 0; i < 16; i++)	// Show the data in Hex:
			{
				if (i && (i&3) == 0)
					putchar(' ');
				if (i < r)
					printf("%02X", buf[i]&0xFF);
				else
					printf("  ");
			}

			// Show the data in Ascii:
			putchar('\t');
			for (i = 0; i < r; i++)
				printf("%c", isprint(buf[i]) ? buf[i] : '.');
			printf("\n");
			prev = 1;
		}

		if (r != thiscount)
			break;		// Short read, abort
		count -= thiscount;
		memcpy(last, buf, thiscount);
	}
}

/*
 * Interactive verify command handler.
 * Verify memory integrity and display errors found.
 * This routine courtesy of Yann Gouy, thanks Yann!
 */
static	void
DebugVerify(char**& argv)
{
	long		debut = eaddr;
	long		compte = 1;
	unsigned int	start;
	int		count;
	int		val_read, val_write;
	uchar*		buf;
	unsigned int	pattern_index;
	int		buf_index;
	uchar		pattern;
	boolean		error;

	GetAddressRange(argv, &debut, &compte);
	start = (unsigned int)debut; count = (unsigned int)compte;
	printf("Verifying memory from 0x%04x to 0x%04x (length = 0x%04x)\n", start, start + count, count);

	/*
	 * Don't set the address if it's set already.
	 */
	if (eaddr != start)
		TalkerSetAddr(start);

	/*
	 * Allocate buffer memory
	 */
	buf = new uchar[count];
	if (buf == NULL)
	{
		printf("Not able to get a %d (0x%04x) byte(s) buffer\n", count, count);
		return;
	}

	/* 
	 * first, write simple patterns to check short-circuits
	 */
	for (pattern_index = 0; pattern_index <= 8; pattern_index++)
	{
		pattern = 1 << pattern_index;
		printf("pattern 0x%02x :\n", pattern);
		error = FALSE;

		// init of the buffer
		for (buf_index = 0; buf_index < count; buf_index++)
		{
			buf[buf_index] = pattern;
		}

		printf(" writing...\n");
		// reset memory address to start
		TalkerSetAddr(start);

		// now, buffer is sent
		val_write = TalkerPutMem(buf, count);
		if (val_write != count)
		{
			printf("Only %d (0x%04x) bytes written out of %d (0x%04x)\n", val_write, val_write, count, count);
		}

		// reset memory address to start
		TalkerSetAddr(start);

		printf(" reading...\n");
		// reread memory but only the number of bytes that was set
		val_read = TalkerGetMem(buf, val_write);
		if (val_read != val_write)
		{
			printf("Only %d (0x%04x) bytes read out of %d (0x%04x)\n", val_read, val_read, val_write, val_write);
		}

		// testing what we get
		printf(" testing...\n");
		for (buf_index = 0; buf_index < count; buf_index++)
		{
			if (buf[buf_index] != pattern)
			{
				fprintf(stderr, "error at 0x%04x : 0x%02x instead of 0x%02x\n", start + buf_index, buf[buf_index], pattern);
				error = TRUE;
			}
		}

		if (error == TRUE)
		{
			printf("error(s) found\n");
		}
	}

	/*
	 * then we write address depend patterns
	 */
	printf("address dependent patterns :\n");

	error = FALSE;

	// init of the buffer
	for (buf_index = 0; buf_index < count; buf_index++)
	{
		buf[buf_index] = ((uchar)(start + buf_index) + (uchar)((start + buf_index) >> 8)) & 0xff;
	}

	printf(" writing...\n");
	// reset memory address to start
	TalkerSetAddr(start);

	// now, buffer is sent
	val_write = TalkerPutMem(buf, count);
	if (val_write != count)
	{
		printf("Only %d (0x%04x) bytes written out of %d (0x%04x)\n", val_write, val_write, count, count);
	}

	// reset memory address to start
	TalkerSetAddr(start);

	printf(" reading...\n");
	// reread memory but only the number of bytes that was set
	val_read = TalkerGetMem(buf, val_write);
	if (val_read != val_write)
	{
		printf("Only %d (0x%04x) bytes read out of %d (0x%04x)\n", val_read, val_read, val_write, val_write);
	}

	// testing what we get
	printf(" testing...\n");
	for (buf_index = 0; buf_index < count; buf_index++)
	{
		if (buf[buf_index] != ((uchar)(start + buf_index) + (uchar)((start + buf_index) >> 8) & 0xff))
		{
			fprintf(stderr, "error at 0x%04x : 0x%02x instead of 0x%02x\n", start + buf_index, buf[buf_index], (uchar)(((start + buf_index) + (start + buf_index) >> 8) & 0xff));
			error = TRUE;
		}
	}

	if (error == TRUE)
	{
		printf("Error(s) found\n");
	}

	/*
	 * freeing buffer
	 */
	delete [] buf;
	
	printf("Finished!\n");
}

/*
 * Interactive read command handler.
 * Read memory and disassemble it.
 */
static	void
DebugDisassemble(char**& argv)
{
	long		start;			// Address and count
	long		count = 1;

	start = eaddr;
	GetAddressRange(argv, &start, &count);

	Disassemble(start, start+count);
}

/*
 * Zero the memory in an address range
 */
static	void
DebugZero(char**& argv)
{
	long		start = eaddr, count = 1;
	uchar		buf[64];

	GetAddressRange(argv, &start, &count);

	memset(buf, 0, sizeof(buf));
	while (count > 64)
	{
		/*
		 * The "FALSE" in the next line assumes that EEPROM erases to
		 * 0xFF. This is true for all 68HC11's and perhaps all EEPROMs.
		 */
		if (TalkerWrite(start, buf, 64, FALSE) != 64)
			return;
		count -= 64;
		start += 64;
	}
	TalkerWrite(start, buf, count, TRUE);
}

static	void
DebugLoadS19(char**& argv)
{
	char*		file;
	FILE*		fp;
	S19		srec;
	int		nextaddr = -1;	// No such address
	int		count = 0;
	int		done = 0;

	if (!(file = *argv)
	 || *file == ';'
	 || *file == '#')
	{
		printf("Load what file?\n");
		return;
	}
	else
		argv++;

	fp = PathOpen(file, "r", FALSE);
	if (!fp)
	{
		perror(file);
		return;
	}
	printf("Loading file %s\n", file);

	while (ReadS19(fp, &srec))
	{
		if (srec.bytes == 0)
			continue; /* Don't be bothered with empty records */
		if (srec.address != nextaddr)
		{
			if (done || count)
				printf("%d (0x%x) bytes of %d (0x%x)\n", done, done, count, count);
			done = count = 0;
			nextaddr = srec.address;
			printf("Loading to 0x%04X, ", srec.address);
			fflush(stdout);
		}
		else if (verbose)
			printf("%d to 0x%04X, ", srec.bytes, srec.address);
		nextaddr += srec.bytes;
		count += srec.bytes;
#if 1
		done += TalkerWrite(
				srec.address, (uchar*)srec.data,
				srec.bytes, TRUE
			);
#else
		done += srec.bytes;
#endif
	}
	if (done || count)
		printf("%d (0x%x) bytes of %d (0x%x)\n", done, done, count, count);

	fclose(fp);
}

/*
 * Get the registers and display them.
 */
static void
DebugRegisters(char**& /* argv */)
{
	Frame		regs;

	memset(&regs, 0xFF, sizeof(regs));
	TalkerGetRegisters(&regs);
	PrintRegisters(&regs);
}

/*
 * Set the specified register
 */
static void
DebugSet(char**& argv)
{
	Frame		regs;
	char*		cp;
	char*		ep = 0;		// Points to byte after converted value
	char*		ap = 0;		// Points to byte after converted addr
	long		val;
	long		addr;
	char*		expr;

	if (!(expr = *argv)
	 || *expr == ';'
	 || *expr == '#')
	{
		printf("Set what?\n");
		return;
	}
	else
		argv++;

	// REVISIT: Need to handle spaces (multiple words) in set commands

	if (!*argv || **argv != '=')
	{
		printf("Can't show yet\n");
		return;
	}
	argv++;

	if (!(cp = *argv))
	{
		printf("No value given for set command\n");
		return;
	}

	argv++;
	ep = 0;			// Get value
	val = strtol(cp, &ep, 0);
	if (ep == cp)
	{
		printf("Set %s to what?\n", expr);
		return;
	}

	/*
	 * The stack is saved here, not in the MCU, ready for "Revert" command.
	 */
	if (strcasecmp(expr, "s") == 0
	 || strcasecmp(expr, "sp") == 0)
	{
		TalkerSetStack(val);
		TalkerGetRegisters(&regs);
		PrintRegisters(&regs);
		return;
	}

	memset(&regs, 0xFF, sizeof(regs));
	TalkerGetRegisters(&regs);

	switch (tolower(*expr))
	{
	case 'a':
		if (strcasecmp(expr, "a") != 0)
			goto unknown;
		regs.acca = val;
		break;

	case 'b':
		if (strcasecmp(expr, "b") != 0)
			goto unknown;
		regs.accb = val;
		break;

	case 'c':
		if (strcasecmp(expr, "c") != 0
		 && strcasecmp(expr, "ccr") != 0)
			goto unknown;
		regs.ccr = val;
		break;

	case 'd':
		if (strcasecmp(expr, "d") != 0)
			goto unknown;
		regs.acca = val>>8;
		regs.accb = val;
		break;

	case 'p':
		if (strcasecmp(expr, "p") != 0
		 && strcasecmp(expr, "pc") != 0)
			goto unknown;
		regs.pc = val;
		break;

	case 'x':
		if (strcasecmp(expr, "x") != 0)
			goto unknown;
		regs.x = val;
		break;

	case 'y':
		if (strcasecmp(expr, "y") != 0)
			goto unknown;
		regs.y = val;
		break;

	case '*':
		addr = strtol(expr+1, &ap, 0);
		if (ap != expr+1)
		{
			char	buf[2];

			cp = buf;
			if (val > 0xFF || *ep == 'w')
				*cp++ = val>>8;
			*cp++ = val;
			val = TalkerWrite(addr, (uchar*)buf, cp-buf, TRUE);
			if (val != cp-buf)
			{
				printf("Write %d bytes to 0x%04lX failed\n",
					cp-buf, addr
				);
			}
		}
		else
			printf("Invalid address %s\n", expr);
		return;

	default:
	unknown:
		// REVISIT: We need to support user-defined symbols here
		printf("Register %s unknown\n", expr);
		return;
	}

	if (!TalkerSetRegisters(&regs))
		printf("Failed to set %s\n", expr);
	TalkerGetRegisters(&regs);
	PrintRegisters(&regs);
}

/*
 * Breakpoint setting and listing:
 *
 * b l		- list breakpoints
 * b .		- set breakpoint at current PC
 * b (address)	- set breakpoint at address
 */
static void
DebugBreakpoint(char**& argv)
{
	char*		cp;
	int		brknum;

	if (!(cp = *argv)
	 || *cp == 'l'
	 || *cp == 'L')
	{
		BreakList();
		if (cp)
			argv++;
		return;
	}
	argv++;

	if (*cp == '.')
	{
		Frame	regs;
		TalkerGetRegisters(&regs);
		if ((brknum = BreakAdd(regs.pc)) >= 0
		 && !QUIET_MODE)
			printf("Set breakpoint %d at 0x%X\n", brknum, regs.pc);
		return;
	}
	else
	{
		char*	ep;
		int	brkaddr = strtol(cp, &ep, 0);	// Is it an address?

		if (ep > cp)
		{
			if ((brknum = BreakAdd(brkaddr)) >= 0
			 && !QUIET_MODE)
				printf(
					"Set breakpoint %d at 0x%X\n",
					brknum,
					brkaddr
				);
			return;
		}
	}
	printf("breakpoint %s?\n", cp);
}

static void
DebugEnable(char**& argv, boolean enable)
{
	char*		cp;
	boolean		all = FALSE;

	if ((cp = *argv) != 0)
		argv++;
	if (!cp
	 || *cp == '*')
		all = TRUE;

	if (all)
	{
		for (int i = BreakCount()-1; i >= 0; i--)
			(void)BreakEnable(i, enable);
		printf(
			"All breakpoints %sabled\n",
			enable ? "en" : "dis"
		);
	}
	else
	{
		char*	ep;
		int	brknum = strtol(cp, &ep, 0);	// Is it an address?

		if (ep > cp)
		{
			if (!BreakEnable(brknum, enable))
				printf("Can't enable breakpoint %d\n", brknum);
		}
		else
			printf("No breakpoint %s\n", cp);
	}
}

/*
 * Remove a breakpoint
 */
static void
DebugUnbreak(char**& argv)
{
	char*		cp;
	boolean		all = FALSE;

	if ((cp = *argv) != 0)
		argv++;
	if (!cp
	 || *cp == '*')
		all = TRUE;

	if (all)
	{
		stopped_at = -1;
		for (int i = BreakCount()-1; i >= 0; i--)
			(void)BreakRemove(i);
		printf("All breakpoints removed\n");
	}
	else
	{
		char*	ep;
		int	brknum = strtol(cp, &ep, 0);	// Is it an address?

		if (ep > cp)
		{
			if (stopped_at == brknum)
				stopped_at = -1;
			(void)BreakRemove(brknum);
			printf("breakpoint %d removed\n", brknum);
		}
		else
			printf("No breakpoint %s\n", cp);
	}
}

static void
DebugStop(Frame* regs)
{
	if ((stopped_at = BreakSetAt(regs->pc)) != -1)
		printf("Stopped at breakpoint %d\n", stopped_at);
	else
		/*
		 * Must be a user-programmed breakpoint from an interrupt
		 * handler. At least, we have to hope that there's an
		 * interrupt frame where we read registers from!
		 */
		printf("\nProgrammed stop at \n");

	Disassemble(regs->pc, regs->pc+1);
	PrintRegisters(regs);

#if 0		/* It's useful to enable this for testing branch prediction */
	ushort	over_call;
	ushort	next = NextInstruction(regs, &over_call);
	printf(
		"Next instruction will %s to 0x%04X\n",
		over_call != 0xFFFF ? "call" : "proceed",
		next
	);
#endif
}

/*
 * Continue command.
 * No parameters yet, but could be:
 * - run to given address
 * - resume from given address
 * - skip N breakpoints
 * - ? probably other alternatives
 */
static void
DebugContinue(
	char**& // argv
)
{
	Frame		regs;

	/*
	 * Single-step if we were stopped at a breakpoint.
	 */
	if (stopped_at >= 0
	 && SingleStep(&regs, FALSE))
	{
		DebugStop(&regs);
		return;
	}

	stdio->disableSpecialInput();		// Raw mode
	TalkerRun();
	stdio->disableSpecialInput(FALSE);

	/*
	 * The instruction at the current PC will be the instruction
	 * *after* the one that caused the interrupt.
	 *
	 * If this location follows a breakpoint we've placed, we:
	 *  - back up one
	 *  - disassemble the instruction that would have been executed instead.
	 *  - remember which breakpoint we stopped at so we can continue.
	 */
	TalkerGetRegisters(&regs);
	if ((stopped_at = BreakSetAt(regs.pc-1)) != -1)
	{
		regs.pc--;
		TalkerSetRegisters(&regs);
	}

	DebugStop(&regs);
}

static void
DebugSingleStep(boolean over)
{
	Frame regs;

	if (SingleStep(&regs, over))
		DebugStop(&regs);
	else
	{
		Disassemble(regs.pc, regs.pc+1);
		// PrintRegisters(&regs);
	}
}

/*
 * We must disable any breakpoints at the current address,
 * do our single step, then restore the breakpoints.
 */
static boolean
SingleStep(Frame* regs, boolean over)
{
	TalkerGetRegisters(regs);

	boolean	was_interruptable = (regs->ccr&CCR_I) == 0;
	ushort	over_call;
	ushort	next_instr = NextInstruction(regs, &over_call);
	int	step_break;
	int	breaks_disabled[16];
	int	num_disabled;
	int	i;

	/*
	 * Disable all active breakpoints at current address
	 */
	for (num_disabled = 0; num_disabled < 16; num_disabled++)
	{
		if ((i = BreakSetAt(regs->pc)) < 0)
			break;
		breaks_disabled[num_disabled] = i;
		BreakEnableQuiet(i, FALSE);
	}

	if (was_interruptable)
	{				// Mask interrupts
		regs->ccr |= CCR_I;
		TalkerSetRegisters(regs);
	}
	if (over && over_call != 0xFFFF)
		next_instr = over_call;
	step_break = BreakAdd(next_instr);

	TalkerRun();

	/*
	 * Find out where we stopped:
	 */
	TalkerGetRegisters(regs);
	if (regs->pc != next_instr+1)
		printf("Single-step ran away to 0x%04X!\n", regs->pc-1);
	regs->pc--;

	// Reinstate interrupts if necessary
	if (was_interruptable)
		regs->ccr &= ~CCR_I;
	TalkerSetRegisters(regs);

	BreakRemove(step_break);	// remove the temporary breakpoint

	// Reenable the breakpoints we passed:
	for (i = 0; i < num_disabled; i++)
		BreakEnableQuiet(breaks_disabled[i], TRUE);

	stopped_at = BreakSetAt(regs->pc);
	return stopped_at >= 0;
}

static void
PrintRegisters(Frame* regs)
{
	printf(
		"sp=0x%04X, ccr=0x%02X, d=0x%04X"
		" (a=0x%02X, b=0x%02X), x=0x%04X, y=0x%04X, pc=0x%04X\n",
		regs->sp, regs->ccr, (regs->acca<<8)+regs->accb,
		regs->acca, regs->accb, regs->x, regs->y, regs->pc
	);
}

/*
 * Disassembly support code
 */
typedef struct
{
	ulong	pc;
	uchar	buf[16];
	int	count;
	int	used;
	uchar	extra;
} DisState;

static int
DisGetByte(void* closure)
{
	DisState*	ds = (DisState*)closure;

	if (ds->pc > 0xFFFF)
		return EOF;

	/* Buffer need filling? */
	if (ds->used >= ds->count)
	{
		int	ask = 0x10000 - ds->pc;	// Bytes to ask for

		/* Don't set the address if it's set already. */
		if (eaddr != ds->pc)
			TalkerSetAddr(ds->pc);
		ds->count = 0;
		if (ask > 16)
			ask = 16;		// Be realistic!
		ds->count = TalkerGetMem(ds->buf, 16);
		ds->used = 0;

		/*
		 * Use real opcodes for any breakpoints in this address range
		 */
		for (ushort p = ds->pc; p < ds->pc+ds->count; p++)
		{
			int	brknum = BreakSetAt(p);
			if (brknum < 0)
				continue;
			ds->buf[p - ds->pc] = BreakOpCode(brknum);
		}
	}
	if (ds->used >= ds->count)
		return EOF;		// failed.
	return ds->buf[ds->used++];	// yes.
}

static void
DisUngetByte(void* closure, char /* ch */)
{
	DisState*	ds = (DisState*)closure;

	ds->used--;	/* Ignore ch, it must be buf[used-1] */
}

static const char*
DisSymbolNear(
	ulong addr,
	ulong* actual_p,
	boolean is_data
)
{
	static	char*	name;
	SymState	state;
	ulong		actual;

	if (name)
	{
		delete name;
		name = 0;
	}
	state = SymbolNear(addr, name, is_data ? SymData : SymCode);
	if (state == SymUndef)
		return 0;
	(void)SymbolValue(name, actual);

	if (addr-actual > 1024)
		return 0;
	if (is_data && actual != addr)
		return 0;		// Only match exact data address
	if (actual_p)
		*actual_p = actual;	// Return actual address
	return name;
}

static void
Disassemble(ushort start, ushort finish)
{
	DisState	ds;
	char*		cp = 0;

	ds.pc = start;
	ds.count = 0;
	ds.used = 0;
	ds.extra = 0;

	while (ds.pc < finish
		&& (cp = disas1(
				DisGetByte,	// get another byte
				DisUngetByte,	// unget one byte
				&ds,		// Value to pass to get/unget
				&ds.pc,		// Program counter
				0,		// Disassembly flags
				&ds.extra,	// Extra flag (can pass 0)
				0,		// Branch prediction not needed
				DisSymbolNear	// Symbol lookup
			)
		)
	)
		printf("%s\n", cp);

	// Prepare for next disassemble command:
	TalkerSetAddr(ds.pc);
}

static uchar
NextGet1(ushort addr)
{
	uchar	buf;
	int	r;

	if (eaddr != addr)
		TalkerSetAddr(addr);

	if ((r = TalkerGetMem(&buf, 1)) != 1)
		printf(
			"Can't fetch memory at 0x%04X, can't predict branch\n",
			addr
		);
	return buf;
}

static ushort
NextInstruction(Frame* regs, ushort* over_call)
{
	DisState	ds;		/* Disassembly state structure */
	Predict		predict;	/* Prediction structure */

	ds.pc = regs->pc;
	ds.count = 0;
	ds.used = 0;
	ds.extra = 0;

	predict.ccr = regs->ccr;		/* Condition codes */
	predict.pc = regs->pc;			/* Current program counter */
	predict.sp = regs->sp;			/* Stack pointer */
	predict.x = regs->x;			/* X index register */
	predict.y = regs->y;			/* Y index register */
	predict.getmem = NextGet1;		/* Get memory byte */

	(void)disas1(
			DisGetByte,	/* get another byte */
			DisUngetByte,	/* unget one byte */
			&ds,		/* Value to pass to get/unget */
			&ds.pc,		/* Program counter */
			0,		/* Disassembly flags */
			&ds.extra,	/* Extra flag (can pass 0) */
			&predict
		);
	if (over_call)
		*over_call = predict.call ? ds.pc : 0xFFFF;
	return predict.branch_to;		/* Out: Next pc */
}
