/*
 * $Id: disasm.h,v 1.4 2001/02/06 11:14:55 cjh Exp $
 *
 * Disassemble HC11 Instructions.
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#if	!defined(DISASM_HXX)
#define	DISASM_HXX

#define	DISA_NOADDR	0x1		/* Don't include code address */
#define	DISA_NOBINARY	0x2		/* Don't include binary code */

/*
 * This structure is not required for disassembly, but
 * you must pass it if you want branch prediction.
 */
typedef struct
{
	uchar	ccr;				/* Condition codes */
	ushort	pc;				/* Current program counter */
	ushort	sp;				/* Stack pointer */
	ushort	x;				/* X index register */
	ushort	y;				/* Y index register */
	uchar	(*getmem)(ushort addr);		/* Get memory byte */
	boolean	call;				/* Is it a call? */
	ushort	branch_to;			/* Out: Next pc */
} Predict;

/*
 * Disassemble one instruction.
 * The data bytes will be returned in sequence by getbyte(), though one
 * extra byte can be pushed back by ungetbyte() (the byte is passed as
 * a convenience in case the caller doesn't still have it). These two
 * functions also receive the closure value, which is otherwise unused.
 *
 * The "extra" pointer may point to a flag byte that can maintain context
 * from one call to the next - info like "the last instruction was an
 * unconditional jump and you can add a blank line". The value pointed to
 * should be initialised to zero unless still set from a previous call.
 *
 * The Predict structure should contain current values of significant
 * registers and a get-memory function, and the branch_to member will
 * be filled in with the location of the next instruction.  It's used
 * during single-stepping.
 *
 * The symbol() function should return the name of a symbol of the specified
 * type which is at or prior to the given address.
 */
char*	disas1(
		int	(*getbyte)(void* closure), /* get another byte */
		void	(*ungetbyte)(void* closure, char), /* unget one byte */
		void*	closure,		/* Value to pass to get/unget */
		ulong*	pc_p,			/* Program counter */
		int	flags,			/* Disassembly flags */
		uchar*	extra = 0,		/* Extra flag */
		Predict* predict = 0,		/* Branch prediction */
		const char* (*symbol)(ulong addr, ulong* actual, boolean is_data) = 0
	);

#endif	/* DISASM_HXX */
