/*
 * $Id: map.h,v 1.4 2002/03/13 11:54:58 cjh Exp $
 *
 * Memory map for the selected chip.
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#if	!defined(MAP_H)
#define	MAP_H

/*
 * This structure defines how to program a type of memory.
 * If prog_reg is zero, the rest of the structure is ignored.
 *
 * This entire structure will be replaced shortly, don't rely on it.
 */
struct MemType
{
	const char*	name;		// Memory type name

	ushort		prog_reg;	// Programming register addr, 0 for RAM
	int		delay;		// Milliseconds for programming
	boolean		external_voltage; // Need external programming voltage?

	uchar		prep_prog;	// Prepare to program byte
	uchar		start_prog;	// Program byte
	uchar		end_prog;	// Finish programming byte

	uchar		prep_erase_all;	// Prepare to erase whole block
	uchar		start_erase_all;// Erase block
	uchar		end_erase_all;	// Finish erasing block

	uchar		prep_erase;	// Prepare to byte-erase
	uchar		start_erase;	// Erase byte
	uchar		end_erase;	// Finish erasing byte
};

/*
 * Structure to define a contiguous block of a single type of memory.
 */
typedef	struct
{
	ushort		start_addr;	// Inclusive
	ushort		end_addr;	// Exclusive
	MemType*	mem_type;	// What kind of memory?

	ushort		Count() const
			{ return (end_addr ? end_addr : 0x10000) - start_addr; }
} MemRegion;

/*
 * Methods to manipulate the machine maps
 */
boolean		MachineShowAll();	// List all machine names
boolean		MachineSelect(		// Select or create a machine
			char**& argv
		);
boolean		MapCommand(		// Add or modify a map entry
			char**& argv
		);
MemRegion const* MapAddr(		// What region contains addr?
			ushort addr
		);
boolean		RegionShowAll(		// Display regions in machine (or cur)
			char* machine_name = 0
		);

#endif	/* MAP_H */
