/*
 * $Id: multiplex.cxx,v 1.6 2002/03/13 10:17:19 cjh Exp $
 *
 * Inout source multiplexor
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#include	<stdincs.h>
#include	<util.h>
#include	<multiplex.h>

/*
 * Poll the array of Multiplexed objects, tell whichever ones are ready,
 * and return TRUE if any were (in the allowed time)
 */
boolean
Multiplex(
	Multiplexed**	mpx_pp,
	int		num_mpx
)
{
	fd_set		rmask;
	fd_set		wmask;
	fd_set		emask;
	int		i;
	struct	timeval	tv;
	long		max_delay = 3600000;	// One hour.
	int		num_ready = FALSE;
	int		max_fd = -1;
	static	int*	fds;
	static	int	num_fds;

	/*
	 * Poll the Multiplexed objects:
	 * - save the file descriptors
	 * - set up select() masks
	 */
	if (num_fds < num_mpx)
	{
		delete [] fds;
		fds = new int[num_fds = num_mpx];
	}
	FD_ZERO(&rmask);
	FD_ZERO(&wmask);
	FD_ZERO(&emask);
	for (i = 0; i < num_mpx; i++)
	{
		boolean	is_output;
		long	delay;

		fds[i] = -1;	// Assume no request
		is_output = FALSE;
		delay = -1;
		if (!mpx_pp[i]->request(fds+i, &is_output, &delay))
			continue;

		if (max_fd < fds[i])
			max_fd = fds[i];

		if (delay >= 0
		 && max_delay > delay)
			max_delay = delay;

		FD_SET(fds[i], is_output ? &wmask : &rmask);
	}

	// TraceTime("Selecting from %d sources (maxfd %d) timeout %ld\n", num_mpx, max_fd, max_delay);
	int	n;

	/*
	 * Select the input sources for the maximum time
	 */
	tv.tv_sec = max_delay/1000;
	tv.tv_usec = (max_delay%1000)*1000;
	// printf("tv_sec %ld, tv_usec %ld\n", tv.tv_sec, tv.tv_usec);
	while ((n = select(max_fd+1, &rmask, &wmask, &emask, &tv)) < 0)
	{
		if (errno == EINTR)
			continue;	/* Error, stopped */
		return FALSE;		/* Other error */
	}
	// TraceTime("Selected, %d sources ready\n", n);

	/*
	 * Announce readiness to whichever are
	 */
	num_ready = 0;
	for (i = 0; i < num_mpx; i++)
	{
		if (fds[i] < 0)
			continue;	// No request
		if (FD_ISSET(fds[i], &rmask)
		 || FD_ISSET(fds[i], &wmask)
		 || FD_ISSET(fds[i], &emask))
		{
		 	mpx_pp[i]->ready();
			num_ready++;
		}
	}
	// printf("%d sources are ready\n", num_ready);

	return num_ready > 0;
}
