/*
 * $Id: serial.h,v 1.8 2002/03/16 11:07:20 cjh Exp $
 *
 * Serial device (tty) handler
 *
 * Copyright (c) 1996-2000, Clifford Heath.
 *
 * This program may be used for any commercial or noncommercial purpose
 * without charge, but acknowledgement of authorship and this copyright
 * message must be retained in any copies.
 */
#ifndef	SERIAL_HXX
#define	SERIAL_HXX

#include	<multiplex.h>
#if defined(__linux)
#include	<sys/ioctl.h>
#include	<linux/serial.h>
#endif

class	Serial : public Multiplexed
{
public:
	virtual		~Serial();
			Serial(const char* tty = 0);

	/* Input modes: */
	virtual	void	disableLineEdit(boolean disabled = TRUE);
	virtual	void	disableSpecialInput(boolean disabled = TRUE);
	virtual	size_t	read(unsigned char* buf, size_t max, size_t min = 1);
	virtual	void	discard();	// Discard unread input

	/* Output modes: */
	virtual	void	disableSpecialOutput(boolean disabled = TRUE);
	virtual	void	write(const unsigned char* buf, size_t count);
	virtual	void	drain();	// Wait until output drains

	virtual	void	suspend();	// suspend special modes
	virtual	void	resume();	// resume special modes

	virtual	boolean	setSpeed(int baud);
	virtual	int	getSpeed();
	virtual long	setMaxDelay(long millis = -1);
	virtual	boolean	isReady() { return is_ready; }

	virtual	boolean	isOk();		// Did constructor succeed?

protected:

#if defined(__unix)
	int		fd;
	struct termios	orig;		// tty state structure
	struct termios	mode;		// tty state structure
#if defined(__linux)
	struct serial_struct	ss_orig; // Special serial IO modes
	struct serial_struct	ss_mode; // Special serial IO modes
#endif
#elif defined(WIN32)
	HANDLE		comms;		// Handle to open comm port
	OVERLAPPED	read_io;	// Async I/O completion structures
	OVERLAPPED	write_io;
	DCB		orig;		// comport state
	DCB		mode;		// comport state
#endif
	int		baud_rate;	// baud rate
	boolean		is_ready;	// Is it ready?
	long		max_delay;	// in milliseconds

private:
	/* functions inherited from Multiplexed: */
	virtual	boolean	request(	// Is a request required?
				int*		fd,		// fd to poll
				boolean*	is_output,	// read or write
				long*		max_delay	// in millisecs
			);
	virtual	void	ready();	// Multiplexer found it ready
};

#endif	/* SERIAL_HXX */
