;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; $Id: eeprom.s,v 1.10 2001/01/19 01:46:41 cjh Exp $
;
; 68hc11 debug monitor (initialisation of EEPROM program routine)
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Configuration values: You might need to change these.
; Set the address in RAM to where the ROM programming routine will be copied:
Prog1Rom 	= 	STACK_BASE-(Prog1RomEnd-Prog1RomStart)
					; Location for Prog1Rom routine

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Initialisation, copy Prog1ROM function into RAM:
CopyProg1Rom:
		ldx	#Prog1RomStart
		ldy	#Prog1Rom
		ldab	#Prog1RomEnd-Prog1RomStart	; set size
CopyLoop:	ldaa	0,x		; Get byte
		staa	0,y		; Store it
		inx			; Increment pointers
		iny
		decb
		bne	CopyLoop	; Still some bytes to copy

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Include talker shared code
		include "protocol.ext.s"
		include	"progrom.ext.s"	; ProgRom routine, with ext addressing

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Prog1Rom; Called to program 1 ROM cell
; - X contains address to program
; - Y contains address of control register
; In an (E)EPROM version, this routine MUST be in RAM.
Prog1RomStart:
		ldaa	*EPREP		; Get "Prepare" value
		staa	0,y		; Prepare to latch data and address
		stab	0,x		; Here it is...
		ldaa	*ESTART		; Get "Start" value
		staa	0,y		; Turn on programming voltage
Prog1Wait:				; Get idle character into B
		ldab	REGS+SCSR	; Wait for RDRF
		andb	#RDRF		; character available?
		beq	Prog1Wait	; Received yet?
		ldab	REGS+SCDR	; get the character
		beq	Prog1Wait	; Still IDLING?
		ldaa	*EDONE		; Get "Done" value
		staa	0,y		; Turn off programming voltage
		ldab	0,x		; Re-read programmed location
		rts			; Done
Prog1RomEnd:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Variables
EADDR		=	VARS_BASE+0	; address of next byte to program
EREG		=	VARS_BASE+2	; address where ELAT and EPROG live
EPREP		=	VARS_BASE+4	; normally with ELAT set
ESTART		=	VARS_BASE+5	; normally with ELAT and EPROG set
EDONE		=	VARS_BASE+6	; normally with ELAT and EPROG clear
